#include <windows.h>
#include <new>
#include <shobjidl.h>
#include <propsys.h>
#include <propkey.h>
#include <propvarutil.h>
#include "RegisterExtension.h"
#include <stdio.h>

#pragma comment(lib, "propsys.lib")

#define Try(expr) hr = expr; if (!SUCCEEDED(hr)) return hr;

long g_cRefModule = 0;
HINSTANCE g_hInst = NULL;

const PROPERTYKEY PKEY_QMT_NES_Mapper	= {{0x8bb45406, 0xb7f1, 0x425d, 0x9b, 0xcd, 0x4c, 0xc1, 0x9d, 0x48, 0xb0, 0xd4}, 100};
const PROPERTYKEY PKEY_QMT_NES_PRG	= {{0x69d18db5, 0xfcf2, 0x4038, 0xba, 0xc3, 0xfc, 0xac, 0xa4, 0xc6, 0x16, 0xcb}, 100};
const PROPERTYKEY PKEY_QMT_NES_CHR	= {{0x3496abc9, 0x6456, 0x4dc2, 0xa1, 0xc2, 0x27, 0x69, 0x44, 0x3c, 0x6a, 0x4f}, 100};

void DllAddRef()
{
	InterlockedIncrement(&g_cRefModule);
}

void DllRelease()
{
	InterlockedDecrement(&g_cRefModule);
}

class __declspec(uuid("27fcbcd6-3cd7-4759-9f44-f44c11827b74")) CNESColumns :
	public IPropertyStore,
	public IPropertyStoreCapabilities,
	public IInitializeWithStream
{
public:
	CNESColumns() : _cRef(1), _pCache(NULL)
	{
		DllAddRef();
	}

	// IUnknown
	IFACEMETHODIMP QueryInterface(REFIID riid, void **ppv)
	{
		static const QITAB qit[] = {
			QITABENT(CNESColumns, IPropertyStore),
			QITABENT(CNESColumns, IPropertyStoreCapabilities),
			QITABENT(CNESColumns, IInitializeWithStream),
			{ 0 },
		};
		return QISearch(this, qit, riid, ppv);
	}

	IFACEMETHODIMP_(ULONG) AddRef()
	{
		return InterlockedIncrement(&_cRef);
	}

	IFACEMETHODIMP_(ULONG) Release()
	{
		ULONG cRef = InterlockedDecrement(&_cRef);
		if (!cRef)
		{
			delete this;
		}
		return cRef;
	}

	// IPropertyStore
	IFACEMETHODIMP GetCount(DWORD *pcProps);
	IFACEMETHODIMP GetAt(DWORD iProp, PROPERTYKEY *pkey);
	IFACEMETHODIMP GetValue(REFPROPERTYKEY key, PROPVARIANT *pPropVar);
	IFACEMETHODIMP SetValue(REFPROPERTYKEY key, REFPROPVARIANT propVar);
	IFACEMETHODIMP Commit();

	// IPropertyStoreCapabilities
	IFACEMETHODIMP IsPropertyWritable(REFPROPERTYKEY key);

	// IInitializeWithStream
	IFACEMETHODIMP Initialize(IStream *pStream, DWORD grfMode);

private:
	~CNESColumns()
	{
		if (_pCache)
			_pCache->Release();
		DllRelease();
	}

	long _cRef;
	BYTE _hdr[16];
	IPropertyStoreCache *_pCache;
};

// Instantiates a property store object
HRESULT CNESColumns_CreateInstance(REFIID riid, void **ppv)
{
	*ppv = NULL;
	CNESColumns *pNew = new(std::nothrow) CNESColumns;
	HRESULT hr = pNew ? S_OK : E_OUTOFMEMORY;
	if (SUCCEEDED(hr))
	{
		hr = pNew->QueryInterface(riid, ppv);
		pNew->Release();
	}
	return hr;
}

// Accessor methods forward directly to internal value cache
HRESULT CNESColumns::GetCount(DWORD *pcProps)
{
	*pcProps = 0;
	return _pCache ? _pCache->GetCount(pcProps) : E_UNEXPECTED;
}

HRESULT CNESColumns::GetAt(DWORD iProp, PROPERTYKEY *pkey)
{
	*pkey = PKEY_Null;
	return _pCache ? _pCache->GetAt(iProp, pkey) : E_UNEXPECTED;
}

HRESULT CNESColumns::GetValue(REFPROPERTYKEY key, PROPVARIANT *pPropVar)
{
	PropVariantInit(pPropVar);
	return _pCache ? _pCache->GetValue(key, pPropVar) : E_UNEXPECTED;
}

// All of these properties are read-only
HRESULT CNESColumns::SetValue(REFPROPERTYKEY, REFPROPVARIANT)
{
	return STG_E_ACCESSDENIED;
}
HRESULT CNESColumns::Commit()
{
	return STG_E_ACCESSDENIED;
}
HRESULT CNESColumns::IsPropertyWritable(REFPROPERTYKEY)
{
	return S_FALSE;
}

// Initialize populates the internal value cache with data from the specified stream
HRESULT CNESColumns::Initialize(IStream *pStream, DWORD)
{
	HRESULT hr;
	if (_pCache)
		return HRESULT_FROM_WIN32(ERROR_ALREADY_INITIALIZED);

	// create the internal value cache
	Try(PSCreateMemoryPropertyStore(IID_PPV_ARGS(&_pCache)))

	// read the header
	ULONG bytesRead;
	pStream->Read(_hdr, 16, &bytesRead);

	// is the header valid?
	if ((bytesRead != 16) || memcmp(_hdr, "NES\x1A", 4) || ((_hdr[7] & 0x0C) == 0x04) || ((_hdr[7] & 0x0C) == 0x0C))
	{
		Try(_pCache->SetState(PKEY_QMT_NES_Mapper, PSC_NOTINSOURCE))
		Try(_pCache->SetState(PKEY_QMT_NES_PRG, PSC_NOTINSOURCE))
		Try(_pCache->SetState(PKEY_QMT_NES_CHR, PSC_NOTINSOURCE))
		return S_OK;
	}

	// start parsing values out of the header
	WCHAR value[8];
	PROPVARIANT propvarValue = {};

	BOOL NES2 = ((_hdr[7] & 0x0C) == 0x08);

	// Mapper
	if (NES2)
		swprintf(value, 8, L"%i.%i", ((_hdr[6] & 0xF0) >> 4) | (_hdr[7] & 0xF0) | ((_hdr[8] & 0x0F) << 8), (_hdr[8] & 0xF0) >> 4);
	else
		swprintf(value, 8, L"%i", ((_hdr[6] & 0xF0) >> 4) | (_hdr[7] & 0xF0));

	Try(InitPropVariantFromString(value, &propvarValue))
	Try(PSCoerceToCanonicalValue(PKEY_QMT_NES_Mapper, &propvarValue))
	Try(_pCache->SetValueAndState(PKEY_QMT_NES_Mapper, &propvarValue, PSC_NORMAL))
	PropVariantClear(&propvarValue);

	// For NES 2.0 exponent-multiplier notation, because normal numbers would be far too large to display
	const char *suffixes[] = {"B","KB","MB","GB","TB","PB","EB"};

	// PRG Size
	if (NES2)
	{
		// Exponent-Multiplier notation
		if ((_hdr[9] & 0xF) == 0xF)
		{
			int exp = (_hdr[4] >> 2) & 0x3F;
			swprintf(value, 8, L"%i %s", (1 << (exp % 10)) * ((_hdr[4] << 1) & 0x7 | 0x1), suffixes[exp / 10]);
		}
		else
			swprintf(value, 8, L"%i KB", 16 * (_hdr[4] | ((_hdr[9] & 0xF) << 8)));
	}
	else
		swprintf(value, 8, L"%i KB", 16 * (_hdr[4]));

	Try(InitPropVariantFromString(value, &propvarValue))
	Try(PSCoerceToCanonicalValue(PKEY_QMT_NES_PRG, &propvarValue))
	Try(_pCache->SetValueAndState(PKEY_QMT_NES_PRG, &propvarValue, PSC_NORMAL))
	PropVariantClear(&propvarValue);

	// CHRSize
	if (NES2)
	{
		// Exponent-Multiplier notation
		if ((_hdr[9] & 0xF0) == 0xF0)
		{
			int exp = (_hdr[5] >> 2) & 0x3F;
			swprintf(value, 8, L"%i %s", (1 << (exp % 10)) * ((_hdr[5] << 1) & 0x7 | 0x1), suffixes[exp / 10]);
		}
		else
			swprintf(value, 8, L"%i KB", 8 * (_hdr[5] | ((_hdr[9] & 0xF0) << 4)));
	}
	else
		swprintf(value, 8, L"%i KB", 8 * (_hdr[5]));

	Try(InitPropVariantFromString(value, &propvarValue))
	Try(PSCoerceToCanonicalValue(PKEY_QMT_NES_CHR, &propvarValue))
	Try(_pCache->SetValueAndState(PKEY_QMT_NES_CHR, &propvarValue, PSC_NORMAL))
	PropVariantClear(&propvarValue);

	return hr;
}

const WCHAR c_szNESFileExtension[] = L".nes";

HRESULT RegisterSchema(BOOL do_reg)
{
	WCHAR path[MAX_PATH];
	GetModuleFileName(g_hInst, path, MAX_PATH);
	PathRemoveFileSpec(path);
	wcscat_s(path, L"\\NESColumns.propdesc");
	if (do_reg)
		return PSRegisterPropertySchema(L"NESColumns.propdesc");
	else	return PSUnregisterPropertySchema(L"NESColumns.propdesc");
}

HRESULT RegisterHandler()
{
	HRESULT hr;
	const WCHAR c_szPropertyHandlerDescription[] = L"NES Columns";

	// register the property schema
	Try(RegisterSchema(TRUE))

	// register the property handler COM object with the system
	CRegisterExtension re(__uuidof(CNESColumns), HKEY_LOCAL_MACHINE);
	Try(re.RegisterInProcServer(c_szPropertyHandlerDescription, L"Both"))
	Try(re.RegisterInProcServerAttribute(L"ManualSafeSave", TRUE))
	Try(re.RegisterPropertyHandler(c_szNESFileExtension))

	return hr;
}

HRESULT UnregisterHandler()
{
	HRESULT hr;

	CRegisterExtension re(__uuidof(CNESColumns), HKEY_LOCAL_MACHINE);

	// Remove the COM object registration.
	Try(re.UnRegisterObject())

	// Unregister the property handler for the file extension.
	Try(re.UnRegisterPropertyHandler(c_szNESFileExtension))

	// Unregister the property schema, since we're the only ones using it
	Try(RegisterSchema(FALSE))

	return hr;
}

typedef HRESULT (*PFNCREATEINSTANCE)(REFIID riid, void **ppvObject);
class CClassFactory : public IClassFactory
{
public:
	CClassFactory(PFNCREATEINSTANCE pfnCreate) : _cRef(1), _pfnCreate(pfnCreate)
	{
		DllAddRef();
	}

	// IUnknown
	IFACEMETHODIMP QueryInterface(REFIID riid, void ** ppv)
	{
		static const QITAB qit[] =
		{
			QITABENT(CClassFactory, IClassFactory),
			{ 0 }
		};
		return QISearch(this, qit, riid, ppv);
	}

	IFACEMETHODIMP_(ULONG) AddRef()
	{
		return InterlockedIncrement(&_cRef);
	}

	IFACEMETHODIMP_(ULONG) Release()
	{
		long cRef = InterlockedDecrement(&_cRef);
		if (cRef == 0)
		{
			delete this;
		}
		return cRef;
	}

	// IClassFactory
	IFACEMETHODIMP CreateInstance(IUnknown *punkOuter, REFIID riid, void **ppv)
	{
		return punkOuter ? CLASS_E_NOAGGREGATION : _pfnCreate(riid, ppv);
	}

	IFACEMETHODIMP LockServer(BOOL fLock)
	{
		if (fLock)
		{
			DllAddRef();
		}
		else
		{
			DllRelease();
		}
		return S_OK;
	}

private:
	~CClassFactory()
	{
		DllRelease();
	}

	long _cRef;
	PFNCREATEINSTANCE _pfnCreate;
};

// Standard DLL functions
STDAPI_(BOOL) DllMain(HINSTANCE hInstance, DWORD dwReason, void *)
{
	if (dwReason == DLL_PROCESS_ATTACH)
	{
		g_hInst = hInstance;
		DisableThreadLibraryCalls(hInstance);
	}
	return TRUE;
}

STDAPI DllCanUnloadNow()
{
	// Only allow the DLL to be unloaded after all outstanding references have been released
	return (g_cRefModule == 0) ? S_OK : S_FALSE;
}

STDAPI DllGetClassObject(REFCLSID clsid, REFIID riid, void **ppv)
{
	*ppv = NULL;
	HRESULT hr = CLASS_E_CLASSNOTAVAILABLE;

	if (clsid == __uuidof(CNESColumns))
	{
		IClassFactory *pClassFactory = new (std::nothrow) CClassFactory(CNESColumns_CreateInstance);
		hr = pClassFactory ? S_OK : E_OUTOFMEMORY;
		if (SUCCEEDED(hr))
		{
			hr = pClassFactory->QueryInterface(riid, ppv);
			pClassFactory->Release();
		}
	}
	return hr;
}

STDAPI DllRegisterServer()
{
	return RegisterHandler();
}

STDAPI DllUnregisterServer()
{
	return UnregisterHandler();
}